#include <stdio.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>

#define NV_PERIOD 999983
#define STDDEV 0.5
#define RES 300
#define TRIALS 20000
#define MARGIN 0.25

#define RES 50
#define TRIALS 4000

long int nvi = 0, nvd = 1;
double nvs[NV_PERIOD];
#define normalvariate() nvs[(nvi = (nvi + nvd) % NV_PERIOD)]

void init_variates() {
    FILE* fp = fopen("random.dat", "r");
    fread(nvs, sizeof(double), NV_PERIOD, fp);
    fclose(fp);
}

long plucounts[4], appcounts[4], irvcounts[4][4][4], concounts[4][4], borcounts[4];
double ax, ay, bx, by, cx, cy, dx, dy;

void clearcounts() {
    int i, j, k;
    for (i = 0; i < 4; i++) {
        borcounts[i] = plucounts[i] = appcounts[i] = 0;
        for (j = 0; j < 4; j++) {
            concounts[i][j] = 0;
            for (k = 0; k < 4; k++) {
                irvcounts[i][j][k] = 0;
            }
        }
    }
}

void countup(double x, double y, double t) {
    double ra = (x - ax)*(x - ax) + (y - ay)*(y - ay);
    double rb = (x - bx)*(x - bx) + (y - by)*(y - by);
    double rc = (x - cx)*(x - cx) + (y - cy)*(y - cy);
    double rd = (x - dx)*(x - dx) + (y - dy)*(y - dy);
    int ab = (ra <= rb), ac = (ra <= rc), ad = (ra <= rd);
    int ba = (rb <= ra), bc = (rb <= rc), bd = (rb <= rd);
    int ca = (rc <= ra), cb = (rc <= rb), cd = (rc <= rd);
    int da = (rd <= ra), db = (rd <= rb), dc = (rd <= rc);

    plucounts[0] += ab & ac && ad;
    plucounts[1] += ba & bc && bd;
    plucounts[2] += ca & cb && cd;
    plucounts[3] += da & db && dc;

    irvcounts[0][1][2] += ab & bc & cd;
    irvcounts[0][1][3] += ab & bd & dc;
    irvcounts[0][2][1] += ac & cb & bd;
    irvcounts[0][2][3] += ac & cd & db;
    irvcounts[0][3][1] += ad & db & bc;
    irvcounts[0][3][2] += ad & dc & cb;
    irvcounts[1][0][2] += ba & ac & cd;
    irvcounts[1][0][3] += ba & ad & dc;
    irvcounts[1][2][0] += bc & ca & ad;
    irvcounts[1][2][3] += bc & cd & da;
    irvcounts[1][3][0] += bd & da & ac;
    irvcounts[1][3][2] += bd & dc & ca;
    irvcounts[2][0][1] += ca & ab & bd;
    irvcounts[2][0][3] += ca & ad & db;
    irvcounts[2][1][0] += cb & ba & ad;
    irvcounts[2][1][3] += cb & bd & da;
    irvcounts[2][3][0] += cd & da & ab;
    irvcounts[2][3][1] += cd & db & ba;
    irvcounts[3][0][1] += da & ab & bc;
    irvcounts[3][0][2] += da & ac & cb;
    irvcounts[3][1][0] += db & ba & ac;
    irvcounts[3][1][2] += db & bc & ca;
    irvcounts[3][2][0] += dc & ca & ab;
    irvcounts[3][2][1] += dc & cb & ba;

    appcounts[0] += (ra <= t*t);
    appcounts[1] += (rb <= t*t);
    appcounts[2] += (rc <= t*t);
    appcounts[3] += (rd <= t*t);

    concounts[0][1] += ab;
    concounts[0][2] += ac;
    concounts[0][3] += ad;
    concounts[1][0] += ba;
    concounts[1][2] += bc;
    concounts[1][3] += bd;
    concounts[2][0] += ca;
    concounts[2][1] += cb;
    concounts[2][3] += cd;
    concounts[3][0] += da;
    concounts[3][1] += db;
    concounts[3][2] += dc;

    borcounts[0] += ab + ac + ad;
    borcounts[1] += ba + bc + bd;
    borcounts[2] += ca + cb + cd;
    borcounts[3] += da + db + dc;
}

void runsim(int trials, double x, double y, double stddev) {
    int i;
    double sx, sy, t;
    clearcounts();
    for (i = 0; i < trials; i++) {
        sx = x + normalvariate()*stddev;
        sy = y + normalvariate()*stddev;
        countup(sx, sy, exp(normalvariate())*0.5);
    }
}

#define move(loser, d1, d2, d3) { \
    irvcounts[d1][d2][d3] += irvcounts[loser][d1][d2] + \
        irvcounts[d1][loser][d2] + irvcounts[d1][d2][loser]; \
    irvcounts[loser][d1][d2] = irvcounts[d1][loser][d2] = \
        irvcounts[d1][d2][loser] = 0; \
}

int red[15], green[15], blue[15];

void init_colours() {
    int i;
    for (i = 0; i < 16; i++) red[i] = green[i] = blue[i] = 0;
    red[1] = 255;
    red[2] = green[2] = 255;
    green[4] = 255;
    blue[8] = 255;
}

int maxwin(int a, int b, int c, int d) {
    if (a > b && a > c && a > d) return 1;
    if (b > a && b > c && b > d) return 2;
    if (c > a && c > b && c > d) return 4;
    if (d > a && d > b && d > c) return 8;
    return 0;
}

int irvwinner() {
    int total = irvcounts[0][1][2] + irvcounts[0][1][3] +
                irvcounts[0][2][1] + irvcounts[0][2][3] +
                irvcounts[0][3][1] + irvcounts[0][3][2] +
                irvcounts[1][0][2] + irvcounts[1][0][3] +
                irvcounts[1][2][0] + irvcounts[1][2][3] +
                irvcounts[1][3][0] + irvcounts[1][3][2] +
                irvcounts[2][0][1] + irvcounts[2][0][3] +
                irvcounts[2][1][0] + irvcounts[2][1][3] +
                irvcounts[2][3][0] + irvcounts[2][3][1] +
                irvcounts[3][0][1] + irvcounts[3][0][2] +
                irvcounts[3][1][0] + irvcounts[3][1][2] +
                irvcounts[3][2][0] + irvcounts[3][2][1];
    int majority = total/2;
    while (1) {
        int va= irvcounts[0][1][2] + irvcounts[0][1][3] +
                irvcounts[0][2][1] + irvcounts[0][2][3] +
                irvcounts[0][3][1] + irvcounts[0][3][2];
        int vb= irvcounts[1][0][2] + irvcounts[1][0][3] +
                irvcounts[1][2][0] + irvcounts[1][2][3] +
                irvcounts[1][3][0] + irvcounts[1][3][2];
        int vc= irvcounts[2][0][1] + irvcounts[2][0][3] +
                irvcounts[2][1][0] + irvcounts[2][1][3] +
                irvcounts[2][3][0] + irvcounts[2][3][1];
        int vd= irvcounts[3][0][1] + irvcounts[3][0][2] +
                irvcounts[3][1][0] + irvcounts[3][1][2] +
                irvcounts[3][2][0] + irvcounts[3][2][1];
        if (va >= majority || vb >= majority ||
            vc >= majority || vd >= majority) {
            return maxwin(va, vb, vc, vd);
        }
        if (va > 0 && (va < vb || vb == 0) && (va < vc || vc == 0) && (va < vd || vd == 0)) {
            move(0,  1, 2, 3);
            move(0,  1, 3, 2);
            move(0,  2, 1, 3);
            move(0,  2, 3, 1);
            move(0,  3, 1, 2);
            move(0,  3, 2, 1);
        } else if (vb > 0 && (vb < va || va == 0) && (vb < vc || vc == 0) && (vb < vd || vd == 0)) {
            move(1,  0, 2, 3);
            move(1,  0, 3, 2);
            move(1,  2, 0, 3);
            move(1,  2, 3, 0);
            move(1,  3, 0, 2);
            move(1,  3, 2, 0);
        } else if (vc > 0 && (vc < va || va == 0) && (vc < vb || vb == 0) && (vc < vd || vd == 0)) {
            move(2,  0, 1, 3);
            move(2,  0, 3, 1);
            move(2,  1, 0, 3);
            move(2,  1, 3, 0);
            move(2,  3, 0, 1);
            move(2,  3, 1, 0);
        } else if (vd > 0 && (vd < va || va == 0) && (vd < vb || vb == 0) && (vd < vc || vc == 0)) {
            move(3,  0, 1, 2);
            move(3,  0, 2, 1);
            move(3,  1, 0, 2);
            move(3,  1, 2, 0);
            move(3,  2, 0, 1);
            move(3,  2, 1, 0);
        } else {
            return maxwin(va, vb, vc, vd);
        }
    }
}

int borwinner() {
    int va = borcounts[0];
    int vb = borcounts[1];
    int vc = borcounts[2];
    int vd = borcounts[3];
    return maxwin(va, vb, vc, vd);
}

int pluwinner() {
    int va = plucounts[0];
    int vb = plucounts[1];
    int vc = plucounts[2];
    int vd = plucounts[3];
    return maxwin(va, vb, vc, vd);
}

int appwinner() {
    int va = appcounts[0];
    int vb = appcounts[1];
    int vc = appcounts[2];
    int vd = appcounts[3];
    return maxwin(va, vb, vc, vd);
}

int conwinner() {
    int va = (concounts[0][1] >= concounts[1][0] &&
              concounts[0][2] >= concounts[2][0] &&
              concounts[0][3] >= concounts[3][0]);
    int vb = (concounts[1][0] >= concounts[0][1] &&
              concounts[1][2] >= concounts[2][1] &&
              concounts[1][3] >= concounts[3][1]);
    int vc = (concounts[2][0] >= concounts[0][2] &&
              concounts[2][1] >= concounts[1][2] &&
              concounts[2][3] >= concounts[3][2]);
    int vd = (concounts[3][0] >= concounts[0][3] &&
              concounts[3][1] >= concounts[1][3] &&
              concounts[3][2] >= concounts[2][3]);
    return maxwin(va, vb, vc, vd);
}

void makemap(double stddev, int res, int trials,
             int axx, int ayy, int bxx, int byy,
             int cxx, int cyy, int dxx, int dyy) {
    int margin = (int) (MARGIN * res);
    int i, j;
    char basename[100], filename[100];
    FILE* pfp;
    FILE* bfp;
    FILE* afp;
    FILE* ifp;
    FILE* cfp;

    ax = axx*0.01;
    ay = ayy*0.01;
    bx = bxx*0.01;
    by = byy*0.01;
    cx = cxx*0.01;
    cy = cyy*0.01;
    dx = dxx*0.01;
    dy = dyy*0.01;

    printf("mapping %02d,%02d %02d,%02d %02d,%02d %02d,%02d\n",
           axx, ayy, bxx, byy, cxx, cyy, dxx, dyy);

    sprintf(basename, "%02d,%02d_%02d,%02d_%02d,%02d_%02d,%02d",
                      axx, ayy, bxx, byy, cxx, cyy, dxx, dyy);

    sprintf(filename, "%s_bor.ppm", basename);
    bfp = fopen(filename, "w");

    sprintf(filename, "%s_plu.ppm", basename);
    pfp = fopen(filename, "w");

    sprintf(filename, "%s_app.ppm", basename);
    afp = fopen(filename, "w");

    sprintf(filename, "%s_irv.ppm", basename);
    ifp = fopen(filename, "w");

    sprintf(filename, "%s_con.ppm", basename);
    cfp = fopen(filename, "w");

    fprintf(pfp, "P6\n%d %d\n255\n", margin*2 + res, margin*2 + res);
    fprintf(afp, "P6\n%d %d\n255\n", margin*2 + res, margin*2 + res);
    fprintf(bfp, "P6\n%d %d\n255\n", margin*2 + res, margin*2 + res);
    fprintf(ifp, "P6\n%d %d\n255\n", margin*2 + res, margin*2 + res);
    fprintf(cfp, "P6\n%d %d\n255\n", margin*2 + res, margin*2 + res);

    for (i = res + margin - 1; i >= -margin; i--) {
        double y = i/(double) res;
        for (j = -margin; j < res + margin; j++) {
            double x = j/(double) res;
            nvi = random() % NV_PERIOD;
            nvd = (random() % (NV_PERIOD - 1)) + 1;

            runsim(trials, x, y, stddev);
            int p = pluwinner();
            fputc(red[p], pfp);
            fputc(green[p], pfp);
            fputc(blue[p], pfp);
            int b = borwinner();
            fputc(red[b], bfp);
            fputc(green[b], bfp);
            fputc(blue[b], bfp);
            int a = appwinner();
            fputc(red[a], afp);
            fputc(green[a], afp);
            fputc(blue[a], afp);
            int i = irvwinner();
            fputc(red[i], ifp);
            fputc(green[i], ifp);
            fputc(blue[i], ifp);
            int c = conwinner();
            fputc(red[c], cfp);
            fputc(green[c], cfp);
            fputc(blue[c], cfp);
        }
        fflush(pfp);
        fflush(afp);
        fflush(bfp);
        fflush(ifp);
        fflush(cfp);
        fputc('.', stdout);
        fflush(stdout);
    }
    fputc('\n', stdout);
    fclose(pfp);
    fclose(afp);
    fclose(bfp);
    fclose(ifp);
    fclose(cfp);
}

#define RC (random() % 100)

int main(int argc, char* argv[]) {
    srandom(0);
    init_variates();
    init_colours();

    if (argc == 1) {
        makemap(STDDEV, RES, TRIALS, 30, 30, 70, 30, 30, 70, 70, 70);
        makemap(STDDEV, RES, TRIALS, 12, 28, 85, 70, 39, 28, 97, 14);
        exit(1);
    }
    if (argc == 2) {
        makemap(STDDEV, RES, TRIALS, 40, 57, 16, 54, 5, 62, 91, 70);
        makemap(STDDEV, RES, TRIALS, 24, 23, 19, 62, 4, 64, 85, 55);
        makemap(STDDEV, RES, TRIALS, 18, 79, 69, 87, 48, 65, 36, 9);
        exit(1);
    }
    if (argc == 3) {
        makemap(STDDEV, RES, TRIALS, 22, 62, 25, 87, 16, 94, 98, 2);
        makemap(STDDEV, RES, TRIALS, 73, 79, 64, 39, 61, 21, 24, 10);
        makemap(STDDEV, RES, TRIALS, 24, 60, 38, 58, 34, 73, 27, 0);
        exit(1);
    }
    if (argc == 4) {
        makemap(STDDEV, RES, TRIALS, 4, 65, 45, 93, 81, 97, 62, 36);
        makemap(STDDEV, RES, TRIALS, 75, 35, 95, 95, 75, 58, 93, 56);
        makemap(STDDEV, RES, TRIALS, 33, 13, 47, 83, 21, 66, 39, 84);
        exit(1);
    }

    srandom(argc);
    while (1) {
        makemap(STDDEV, RES, TRIALS, RC, RC, RC, RC, RC, RC, RC, RC);
    }
}
