#include <stdio.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>

#define NV_PERIOD 7999993
#define STDDEV 0.5
#define RES 300
#define TRIALS 80000
#define MARGIN 0.25

long int nvi = 0, nvd = 1;
double nvs[NV_PERIOD];
#define normalvariate() nvs[(nvi = (nvi + nvd) % NV_PERIOD)]

void init_variates() {
    FILE* fp = fopen("random.dat", "r");
    fread(nvs, sizeof(double), NV_PERIOD, fp);
    fclose(fp);
}

long plucounts[3], appcounts[3], irvcounts[3][3], concounts[3][3];
double ax, ay, bx, by, cx, cy;

void clearcounts() {
    int i, j;
    for (i = 0; i < 3; i++) {
        plucounts[i] = appcounts[i] = 0;
        for (j = 0; j < 3; j++) irvcounts[i][j] = concounts[i][j] = 0;
    }
}

void countup(double x, double y, double t) {
    double da, db, dc;
    int ab, ac, ba, bc, ca, cb;
    double u = 1.0 - t;

    da = hypot(x - ax, y - ay);
    db = hypot(x - bx, y - by);
    dc = hypot(x - cx, y - cy);
    ab = (da <= db);
    ac = (da <= dc);
    ba = (db <= da);
    bc = (db <= dc);
    ca = (dc <= da);
    cb = (dc <= db);

    plucounts[0] += ab & ac;
    plucounts[1] += ba & bc;
    plucounts[2] += ca & cb;

    irvcounts[0][1] += ab & bc;
    irvcounts[0][2] += ac & cb;
    irvcounts[1][0] += ba & ac;
    irvcounts[1][2] += bc & ca;
    irvcounts[2][0] += ca & ab;
    irvcounts[2][1] += cb & ba;

    if (ab & ac) {
        appcounts[0]++;
        if (db < da*t + dc*u) appcounts[1]++;
        if (dc < da*t + db*u) appcounts[2]++;
    }
    if (ba & bc) {
        appcounts[1]++;
        if (da < db*t + dc*u) appcounts[0]++;
        if (dc < db*t + da*u) appcounts[2]++;
    }
    if (ca & cb) {
        appcounts[2]++;
        if (da < dc*t + db*u) appcounts[0]++;
        if (db < dc*t + da*u) appcounts[1]++;
    }

    concounts[0][1] += ab;
    concounts[0][2] += ac;
    concounts[1][0] += ba;
    concounts[1][2] += bc;
    concounts[2][0] += ca;
    concounts[2][1] += cb;
}

void runsim(int trials, double x, double y, double stddev) {
    int i;
    double sx, sy, t;
    clearcounts();
    for (i = 0; i < trials; i++) {
        sx = x + normalvariate()*stddev;
        sy = y + normalvariate()*stddev;
        do t = normalvariate()*0.25 + 0.5;
        while (t < 0.01 || t > 0.99);
        countup(sx, sy, t);
    }
}

#define move(s1, s2, s3, d1, d2) { \
    irvcounts[d1][d2] += irvcounts[s1][s2]; irvcounts[s1][s2] = 0; \
}

#define bitmax(a, b, c) \
    (a > b ? (a == c ? 5 : a > c ? 1 : 4) : \
     b > a ? (b == c ? 6 : b > c ? 2 : 4) : \
             (a == c ? 7 : a > c ? 3 : 4))

int irvwinner() {
    int total = irvcounts[0][1] + irvcounts[0][2] +
                irvcounts[1][0] + irvcounts[1][2] +
                irvcounts[2][0] + irvcounts[2][1];
    int majority = total/2;
    while (1) {
        int va = irvcounts[0][1] + irvcounts[0][2];
        int vb = irvcounts[1][0] + irvcounts[1][2];
        int vc = irvcounts[2][0] + irvcounts[2][1];
        if (va >= majority || vb >= majority || vc >= majority) {
            return bitmax(va, vb, vc);
        }
        if (va < vb && va < vc) {
            move(0, 1, 2,   1, 2);
            move(0, 2, 1,   2, 1);
            move(1, 0, 2,   1, 2);
            move(2, 0, 1,   2, 1);
        } else if (vb < va && vb < vc) {
            move(1, 0, 2,   0, 2);
            move(1, 2, 0,   2, 0);
            move(0, 1, 2,   0, 2);
            move(2, 1, 0,   2, 0);
        } else if (vc < va && vc < vb) {
            move(2, 0, 1,   0, 1);
            move(2, 1, 0,   1, 0);
            move(0, 2, 1,   0, 1);
            move(1, 2, 0,   1, 0);
        } else {
            return bitmax(va, vb, vc);
        }
    }
}

int pluwinner() {
    int va = plucounts[0];
    int vb = plucounts[1];
    int vc = plucounts[2];
    return bitmax(va, vb, vc);
}

int appwinner() {
    int va = appcounts[0];
    int vb = appcounts[1];
    int vc = appcounts[2];
    return bitmax(va, vb, vc);
}

int conwinner() {
    int va = (concounts[0][1] >= concounts[1][0] &&
              concounts[0][2] >= concounts[2][0]);
    int vb = (concounts[1][0] >= concounts[0][1] &&
              concounts[1][2] >= concounts[2][1]);
    int vc = (concounts[2][0] >= concounts[0][2] &&
              concounts[2][1] >= concounts[1][2]);
    return bitmax(va, vb, vc);
}

void makemap(double stddev, int res, int trials,
             int axx, int ayy, int bxx, int byy, int cxx, int cyy) {
    int margin = (int) (MARGIN * res);
    int i, j;
    double sqrad = (2.0 / res) * (2.0 / res);
    char filename[100];
    FILE* pfp;
    FILE* afp;
    FILE* ifp;
    FILE* cfp;

    nvi = random() % NV_PERIOD;
    nvd = random() % (NV_PERIOD/2) + 1;

    ax = axx*0.01;
    ay = ayy*0.01;
    bx = bxx*0.01;
    by = byy*0.01;
    cx = cxx*0.01;
    cy = cyy*0.01;

    sprintf(filename, "plu_%02d,%02d_%02d,%02d_%02d,%02d.ppm",
                      axx, ayy, bxx, byy, cxx, cyy);
    pfp = fopen(filename, "w");

    sprintf(filename, "app_%02d,%02d_%02d,%02d_%02d,%02d.ppm",
                      axx, ayy, bxx, byy, cxx, cyy);
    afp = fopen(filename, "w");

    sprintf(filename, "irv_%02d,%02d_%02d,%02d_%02d,%02d.ppm",
                      axx, ayy, bxx, byy, cxx, cyy);
    ifp = fopen(filename, "w");

    sprintf(filename, "con_%02d,%02d_%02d,%02d_%02d,%02d.ppm",
                      axx, ayy, bxx, byy, cxx, cyy);
    cfp = fopen(filename, "w");

    printf("%s\n", filename);
    fprintf(pfp, "P6\n%d %d\n255\n", margin*2 + res, margin*2 + res);
    fprintf(afp, "P6\n%d %d\n255\n", margin*2 + res, margin*2 + res);
    fprintf(ifp, "P6\n%d %d\n255\n", margin*2 + res, margin*2 + res);
    fprintf(cfp, "P6\n%d %d\n255\n", margin*2 + res, margin*2 + res);

    for (i = res + margin - 1; i >= -margin; i--) {
        double y = i/(double) res;
        for (j = -margin; j < res + margin; j++) {
            double x = j/(double) res;
            runsim(trials, x, y, stddev);
            int p = pluwinner();
            fputc(p & 1 ? 255 : 0, pfp);
            fputc(p & 2 ? 255 : 0, pfp);
            fputc(p & 4 ? 255 : 0, pfp);
            int a = appwinner();
            fputc(a & 1 ? 255 : 0, afp);
            fputc(a & 2 ? 255 : 0, afp);
            fputc(a & 4 ? 255 : 0, afp);
            int i = irvwinner();
            fputc(i & 1 ? 255 : 0, ifp);
            fputc(i & 2 ? 255 : 0, ifp);
            fputc(i & 4 ? 255 : 0, ifp);
            int c = conwinner();
            fputc(c & 1 ? 255 : 0, cfp);
            fputc(c & 2 ? 255 : 0, cfp);
            fputc(c & 4 ? 255 : 0, cfp);
        }
        fflush(pfp);
        fflush(afp);
        fflush(ifp);
        fflush(cfp);
        fputc('.', stdout);
        fflush(stdout);
    }
    fputc('\n', stdout);
    fclose(pfp);
    fclose(afp);
    fclose(ifp);
    fclose(cfp);
}

#define RC (random() % 100)

int main(int argc, char* argv[]) {
    srandom(0);
    init_variates();

    if (argc == 1) {
        makemap(STDDEV, RES, TRIALS, 50, 99, 7, 25, 93, 25);
        makemap(STDDEV, RES, TRIALS, 54, 47, 77, 64, 13, 10);
        makemap(STDDEV, RES, TRIALS, 40, 20, 89, 78, 72, 60);
        makemap(STDDEV, RES, TRIALS, 89, 89, 51, 29, 70, 55);
    }

    srandom(argc);
    while (1) {
        makemap(STDDEV, RES, TRIALS, RC, RC, RC, RC, RC, RC);
    }
}
